import java.lang.*;
import java.util.*;

public class Lighting {

  static final int NONE = 0 ;
  static final int AMBIENT = 1 ;
  static final int POINT = 2 ;

  private class Light {
    public int type = NONE ;
    public double [] parms ;
    public Light(int t, double [] values) {
      type = t;
      parms = values ;
    }
  }

  List lights ;

  public Lighting() {
    lights = new LinkedList();
  }

  public void addAmbientLight(double ia) {
    double [] v = new double[1];
    v[0] = ia;
    lights.add(new Light(AMBIENT, v));
  }

  public void addLightPoint(
      double x, double y, double z, double w, double id)
  {
    double [] v = new double[5];
    v[0] =  x; v[1] =  y; v[2] =  z; v[3] =  w; v[4] =  id;
    lights.add(new Light(POINT, v));
  }

  public double [] applyLights(
      double  x, double  y, double  z,
      double nx, double ny, double nz,
      double ka, double kd, double ks, double p,
      double ex, double ey, double ez,
      double  r, double  g, double  b) throws Exception
  {
    double [] ret = new double[3] ;

    double coef;
    coef = 0.0 ;

    Iterator it;
    it = lights.iterator();
    while (it.hasNext()){
      Light l = (Light) it.next();
      switch (l.type) {
        case AMBIENT:
          {
            coef += l.parms[0] * ka;
            break;
          }
        case POINT:
          {
            double [] nVec = new double[3];
            nVec[0] = nx; nVec[1] = ny; nVec[2] = nz;
            nVec = MatVec.aXAndBY(1. / MatVec.norm3(nVec), nVec, 0, null ) ;
            double [] lVec = new double[3];
            lVec[0] = l.parms[0] - x;
            lVec[1] = l.parms[1] - y;
            lVec[2] = l.parms[2] - z;
            lVec = MatVec.aXAndBY(1. / MatVec.norm3(lVec), lVec, 0, null ) ;
            double cos_theta = MatVec.dotProd3(nVec,lVec) ;
            if (cos_theta >= 0) {
              double [] vVec = new double[3];
              vVec[0] = ex - x;
              vVec[1] = ey - y;
              vVec[2] = ez - z;
              vVec = MatVec.aXAndBY( 1. / MatVec.norm3(vVec), vVec, 0 , null) ;
              double dot_nv = MatVec.dotProd3(nVec,vVec);
              if (dot_nv >= 0) {
                double [] rVec ;
                rVec = MatVec.aXAndBY( -1. , lVec , 2. * cos_theta , nVec) ;
                rVec = MatVec.aXAndBY(1. / MatVec.norm3(rVec), rVec, 0, null ) ;

                double cos_alpha = MatVec.dotProd3(rVec,vVec);

                coef += l.parms[4] * kd * cos_theta;
                coef += l.parms[4] * ks * Math.pow(cos_alpha, p);
              }
            } 
            break;
          }
        default:
          throw new Exception("Unknown LIGHT type !!!");
      }
    } 
    if (coef < 0) {
      //System.out.println("Coef < 0 :" + coef + "   (" + x+", "+y+", "+z+")");
    }
    ret[0] = coef * r;
    ret[1] = coef * g;
    ret[2] = coef * b;

    return ret ;
  }
}
